import abc
import argparse
import glob
import os
from time import sleep
from typing import List, Iterable, Dict, Union

import watchdog.events
import watchdog.observers
from wai.logging import LOGGING_WARNING

from kasperl.api import MetaFileReader, check_dir
from seppl import AnyData, Plugin
from seppl.io import InfiniteReader
from seppl.placeholders import PlaceholderSupporter, placeholder_list

GLOB_NAME_PLACEHOLDER = "{NAME}"
""" The glob placeholder for identifying other input files. """

EVENT_CREATED = "created"
EVENT_MODIFIED = "modified"
EVENTS = [
    EVENT_CREATED,
    EVENT_MODIFIED,
]

WATCH_ACTION_NOTHING = "nothing"
WATCH_ACTION_MOVE = "move"
WATCH_ACTION_DELETE = "delete"
WATCH_ACTIONS = [
    WATCH_ACTION_NOTHING,
    WATCH_ACTION_MOVE,
    WATCH_ACTION_DELETE,
]

POLLING_TYPE_NEVER = "never"
POLLING_TYPE_INITIAL = "initial"
POLLING_TYPE_ALWAYS = "always"
POLLING_TYPES = [
    POLLING_TYPE_NEVER,
    POLLING_TYPE_INITIAL,
    POLLING_TYPE_ALWAYS,
]


class Handler(watchdog.events.PatternMatchingEventHandler):

    def __init__(self, owner):
        self._extensions = ["*" + x for x in owner.extensions]
        self._events = set(owner.events)
        self.owner = owner
        super().__init__(patterns=self._extensions, ignore_directories=True, case_sensitive=False)

    def on_created(self, event):
        if EVENT_CREATED in self._events:
            if self.owner.polling_type == POLLING_TYPE_ALWAYS:
                self.owner.list_files()
            else:
                self.owner.add_file(event.src_path)

    def on_modified(self, event):
        if EVENT_MODIFIED in self._events:
            if self.owner.polling_type == POLLING_TYPE_ALWAYS:
                self.owner.list_files()
            else:
                self.owner.add_file(event.src_path)


class WatchDir(MetaFileReader, InfiniteReader, PlaceholderSupporter, abc.ABC):

    def __init__(self, dir_in: str = None, dir_out: str = None, check_wait: float = None, process_wait: float = None,
                 action: str = None, extensions: List[str] = None, other_input_files: List[str] = None,
                 max_files: int = None, polling_type: str = None, base_reader: str = None,
                 events: Union[str, List[str]] = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param dir_in: the directory to poll for new files
        :type dir_in: str
        :param dir_out: the directory to move the files to before presenting them to the base reader
        :type dir_out: str
        :param check_wait: the seconds to wait before checking whether any files have been discovered
        :type check_wait: float
        :param process_wait: the seconds to wait before processing the files (e.g., to be fully written to disk)
        :type process_wait: float
        :param action: the action to apply to the input files
        :type action: str
        :param extensions: the list of extensions to poll the directory for
        :type extensions: list
        :param other_input_files: other files that need to be present, glob expression (use placeholder GLOB_NAME_PLACEHOLDER)
        :type other_input_files: list
        :param max_files: the maximum number of files to poll (<1 for no limit)
        :type max_files: int
        :param polling_type: what kind of polling to do
        :type polling_type: str
        :param base_reader: the base reader to use (command-line)
        :type base_reader: str
        :param events: the event(s) to watch for
        :type events: str or list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(base_reader=base_reader, logger_name=logger_name, logging_level=logging_level)
        self.dir_in = dir_in
        self.dir_out = dir_out
        self.check_wait = check_wait
        self.process_wait = process_wait
        self.action = action
        self.extensions = extensions
        self.other_input_files = other_input_files
        self.max_files = max_files
        self.polling_type = polling_type
        self.events = events
        self._inputs = None
        self._current_input = None
        self._base_reader = None
        self._actual_dir_in = None
        self._actual_dir_out = None
        self._observer = None
        self._files = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "watch-dir"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Watches a directory for file changes and presents them to the base reader. " \
            + "The 'polling_type' determines how files are being discovered: " \
            + POLLING_TYPE_NEVER + ": always uses files supplied by the watchdog events; " \
            + POLLING_TYPE_INITIAL + ": does a full poll when first starting and then relies on files form watchdog events; " \
            + POLLING_TYPE_ALWAYS + ": performs and initial poll and whenever the watchdog triggers an event."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--dir_in", type=str, help="The directory to poll; " + placeholder_list(obj=self), required=True)
        parser.add_argument("-o", "--dir_out", type=str, help="The directory to move the files to; " + placeholder_list(obj=self), required=False)
        parser.add_argument("-w", "--check_wait", type=float, help="The number of seconds to wait before checking whether any files were discovered.", required=False, default=0.01)
        parser.add_argument("-W", "--process_wait", type=float, help="The number of seconds to wait before processing the polled files (e.g., waiting for files to be fully written)", required=False, default=0.0)
        parser.add_argument("-a", "--action", choices=WATCH_ACTIONS, help="The action to apply to the input files; 'move' moves the files to --dir_out directory", required=False, default=WATCH_ACTION_MOVE)
        parser.add_argument("-e", "--extensions", type=str, help="The extensions of the files to poll (incl. dot)", required=True, nargs="+")
        parser.add_argument("-O", "--other_input_files", type=str, help="The glob expression(s) for capturing other files apart from the input files; use " + GLOB_NAME_PLACEHOLDER + " in the glob expression for the current name", required=False, default=None, nargs="*")
        parser.add_argument("-m", "--max_files", type=int, help="The maximum number of files in a single poll; <1 for unlimited", required=False, default=-1)
        parser.add_argument("-p", "--polling_type", choices=POLLING_TYPES, help="The type of polling type to perform.", required=False, default=POLLING_TYPE_NEVER)
        parser.add_argument("-E", "--events", choices=EVENTS, help="The events to monitor", required=True, default=None, nargs="+")
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.dir_in = ns.dir_in
        self.dir_out = ns.dir_out
        self.check_wait = ns.check_wait
        self.process_wait = ns.process_wait
        self.action = ns.action
        self.extensions = ns.extensions
        self.other_input_files = ns.other_input_files
        self.max_files = ns.max_files
        self.polling_type = ns.polling_type
        self.events = ns.events

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        if self._base_reader is None:
            return [AnyData]
        else:
            return [self._base_reader.generates()]

    @abc.abstractmethod
    def _available_readers(self) -> Dict[str, Plugin]:
        """
        Return the available readers.

        :return: the reader plugins
        :rtype: dict
        """
        raise NotImplementedError()

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """

        super().initialize()

        if self.check_wait is None:
            self.check_wait = 0.01
        if self.process_wait is None:
            self.process_wait = 0.0
        if self.action is None:
            self.action = WATCH_ACTION_MOVE
        if (self.extensions is None) or (len(self.extensions) == 0):
            raise Exception("No extensions defined for polling!")
        if self.max_files is None:
            self.max_files = -1
        if self.polling_type is None:
            self.polling_type = POLLING_TYPE_NEVER
        if (self.events is None) or (len(self.events) == 0):
            raise Exception("No event(s) specified!")

        # check dirs
        self._actual_dir_in = self.session.expand_placeholders(self.dir_in)
        check_dir(self._actual_dir_in, "Input")
        if self.action == WATCH_ACTION_MOVE:
            self._actual_dir_out = self.session.expand_placeholders(self.dir_out)
            check_dir(self._actual_dir_out, "Output")

    def add_file(self, path: str):
        """
        Adds the path to the internal list of files to process.

        :param path: the path to add
        :type path: str
        """
        self._files.append(path)

    def list_files(self):
        """
        Generates the list of files.
        """

        file_list = []
        self.logger().debug("Start listing files: %s" % self._actual_dir_in)

        try:
            for file_name in os.listdir(self._actual_dir_in):
                file_path = os.path.join(self._actual_dir_in, file_name)

                if os.path.isdir(file_path):
                    continue

                # monitored extension?
                if self.extensions is not None:
                    ext_lower = os.path.splitext(file_name)[1]
                    if ext_lower not in self.extensions:
                        self.logger().debug("%s does not match extensions: %s" % (file_name, str(self.extensions)))
                        continue

                file_list.append(file_path)

                # reached limit for poll?
                if self.max_files > 0:
                    if len(file_list) == self.max_files:
                        self.logger().debug("Reached maximum of %d files" % self.max_files)
                        break

            self.logger().debug("Finished listing files")

        except KeyboardInterrupt:
            return
        except:
            self.logger().exception("Failed listing files!")

        self._files = file_list

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self._files = []

        handler = Handler(self)
        observer = watchdog.observers.Observer()
        observer.schedule(handler, path=self._actual_dir_in, recursive=False)
        observer.start()

        # initial listing
        if self.polling_type in [POLLING_TYPE_INITIAL, POLLING_TYPE_ALWAYS]:
            self.list_files()

        while not self.session.stopped:
            if len(self._files) == 0:
                sleep(self.check_wait)
                continue

            if self.process_wait > 0:
                self.logger().info("Waiting for %s seconds before processing" % str(self.process_wait))
                sleep(self.process_wait)
            result = self._read_files(self._files)

            # delete or move files
            for file_path in self._files:
                if self.action == WATCH_ACTION_DELETE:
                    self.logger().debug("Deleting input: %s" % file_path)
                    os.remove(file_path)
                elif self.action == WATCH_ACTION_MOVE:
                    self.logger().debug("Moving input: %s -> %s" % (file_path, self._actual_dir_out))
                    os.rename(file_path, os.path.join(self._actual_dir_out, os.path.basename(file_path)))
                elif self.action == WATCH_ACTION_NOTHING:
                    pass
                else:
                    raise Exception("Unhandled action: %s" % self.action)

                # other input files?
                if self.other_input_files is not None:
                    for other_input_file in self.other_input_files:
                        other_files = glob.glob(os.path.join(self._actual_dir_in, other_input_file.replace(GLOB_NAME_PLACEHOLDER, os.path.splitext(file_path)[0])))
                        for other_file in other_files:
                            other_path = os.path.join(self._actual_dir_in, other_file)
                            if self.action == WATCH_ACTION_DELETE:
                                self.logger().debug("Deleting other input: %s" % other_path)
                                os.remove(other_path)
                            elif self.action == WATCH_ACTION_MOVE:
                                self.logger().debug("Moving other input: %s -> %s" % (other_path, self._actual_dir_out))
                                os.rename(other_path, os.path.join(self._actual_dir_out, os.path.basename(other_path)))
                            elif self.action == WATCH_ACTION_NOTHING:
                                pass
                            else:
                                raise Exception("Unhandled action: %s" % self.action)

            # reset list
            self._files = []

            for item in result:
                yield item

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return False

    def finalize(self):
        """
        Finishes the reading, e.g., for closing files or databases.
        """
        super().finalize()
        if self._observer is not None:
            self._observer.stop()
            self._observer.join()



import os
from typing import List

from kasperl.api import make_list, StreamWriter
from seppl.placeholders import InputBasedPlaceholderSupporter


class DeleteFiles(StreamWriter, InputBasedPlaceholderSupporter):

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "delete-files"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Deletes the files associate with the file names it receives. Placeholders in the file names get expanded automatically."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [str, list]

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        for item in make_list(data):
            if isinstance(item, list):
                self.write_stream(item)
                return
            path = self.session.expand_placeholders(item)
            if not os.path.exists(path):
                self.logger().warning("File does not exist: %s" % path)
                continue
            if os.path.isdir(path):
                self.logger().warning("Not a file: %s" % path)
                continue
            try:
                self.logger().info("Deleting file: %s" % path)
                os.remove(path)
            except:
                self.logger().error("Failed to delete file: %s" % path, exc_info=True)



import argparse
from typing import List

from wai.logging import LOGGING_WARNING
from seppl import MetaDataHandler, get_metadata, AnyData
from seppl.io import Splitter, BatchFilter
from kasperl.api import flatten_list, make_list

META_SPLIT = "split"
""" the key for storing the split name in the meta-data. """


class SplitRecords(BatchFilter):
    """
    Splits the incoming records into the specified split ratios by setting the 'split' meta-data value.
    """

    def __init__(self, split_ratios: List[int] = None, split_names: List[str] = None, field: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param split_ratios: the ratios for splitting the data (must some up to 100)
        :type split_ratios: list
        :param split_names: the names for the splits (will be stored in the meta-data)
        :type split_names: list
        :param field: the metadata field to store the split in
        :type field: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.split_ratios = split_ratios
        self.split_names = split_names
        self.field = field
        self._splitter = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "split-records"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Splits the incoming records into the specified split ratios by setting a meta-data value (default: %s)." % META_SPLIT

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-r", "--split_ratios", type=int, default=None, help="The split ratios to use for generating the splits (must sum up to 100)", nargs="+")
        parser.add_argument("-n", "--split_names", type=str, default=None, help="The split names to use for the generated splits, get stored in the meta-data under the specified key.", nargs="+")
        parser.add_argument("-f", "--field", type=str, help="The meta-data field to use for storing the split", default=META_SPLIT, required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.split_ratios = ns.split_ratios[:]
        self.split_names = ns.split_names[:]
        self.field = ns.field

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()

        self._splitter = Splitter(self.split_ratios, self.split_names)
        self._splitter.initialize()
        if (self.field is None) or (len(self.field.strip()) == 0):
            self.field = META_SPLIT

    def _output_stats(self):
        """
        Outputs the statistics.
        """
        stats = self._splitter.stats()
        for split in stats:
            self.logger().info("%s: %d" % (split, stats[split]))

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        if self._has_input_changed(update=True):
            if self._splitter.counter() > 0:
                self.logger().info("Input changed, resetting splitter.")
                self._output_stats()
                self._splitter.reset()

        result = []
        for item in make_list(data):
            # get meta data
            meta = get_metadata(item)
            if meta is None:
                if not isinstance(item, MetaDataHandler):
                    raise Exception("Cannot access meta-data for type: %s" % str(type(item)))
                else:
                    meta = dict()
                    item.set_metadata(meta)

            # find split according to schedule
            meta[self.field] = self._splitter.next()
            result.append(item)

        return flatten_list(result)

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()
        self._output_stats()



import argparse
from typing import List

from seppl import AnyData
from seppl.io import BatchFilter
from seppl.placeholders import add_placeholder, InputBasedPlaceholderSupporter
from wai.logging import LOGGING_WARNING


class MetaDataToPlaceholder(BatchFilter, InputBasedPlaceholderSupporter):

    def __init__(self, metadata_key: str = None, placeholder: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param metadata_key: the meta-data key to get the placeholder value from
        :type metadata_key: str
        :param placeholder: the name of the placeholder (without curly brackets)
        :type placeholder: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.metadata_key = metadata_key
        self.placeholder = placeholder

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "metadata-to-placeholder"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Sets the placeholder with the value from the meta-data passing through."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-k", "--metadata_key", type=str, help="The key in the meta-data to get the value for the placeholder from.", default=None, required=False)
        parser.add_argument("-p", "--placeholder", type=str, help="The name of the placeholder, without curly brackets.", default=None, required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.placeholder = ns.placeholder
        self.metadata_key = ns.metadata_key

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.metadata_key is None:
            raise Exception("No meta-data key provided!")
        if self.placeholder is None:
            raise Exception("No placeholder name provided!")

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        meta = data.get_metadata()
        # no meta-data?
        if meta is None:
            self.logger().warning("No meta-data present!")
            return data
        # key not present?
        if self.metadata_key not in meta:
            self.logger().warning("Meta-data key '%s' not present!" % self.metadata_key)
            return data

        value = meta[self.metadata_key]
        self.logger().info("%s -> %s" % (self.placeholder, value))
        add_placeholder(self.placeholder, "from meta-data", False, lambda i: value)
        return data



import abc
import argparse
import logging
from typing import List, Dict, Optional, Union

from wai.logging import LOGGING_WARNING, init_logging, add_logging_level, set_logging_level
from seppl import Plugin, PluginWithLogging, Initializable, split_cmdline, split_args, args_to_objects


class Generator(PluginWithLogging, Initializable, abc.ABC):
    """
    Base class for generators.
    """

    def __init__(self, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the generator.

        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)

    def _check(self) -> Optional[str]:
        """
        Hook method for performing checks.

        :return: None if checks successful, otherwise error message
        :rtype: str
        """
        return None

    @abc.abstractmethod
    def _do_generate(self) -> List[Dict[str, str]]:
        """
        Generates the variables.

        :return: the list of variable dictionaries
        :rtype: list
        """
        raise NotImplementedError()

    def generate(self):
        """
        Generates the variables.

        :return: the list of variable dictionaries
        :rtype: list
        """
        msg = self._check()
        if msg is not None:
            raise Exception(msg)
        return self._do_generate()

    @classmethod
    def parse_generator(cls, cmdline: str, available_generators: Dict[str, Plugin]) -> 'Generator':
        """
        Parses the commandline and returns the generator plugin.

        :param cmdline: the command-line to parse
        :type cmdline: str
        :param available_generators: the generators to use for parsing
        :type available_generators: dict
        :return: the generator plugin
        :rtype: Generator
        """
        result = cls.parse_generators(cmdline, available_generators)
        if len(result) != 1:
            raise Exception("Expected a single generator, but got: %d" % len(result))
        return result[0]

    @classmethod
    def parse_generators(cls, cmdline: str, available_generators: Dict[str, Plugin]) -> List['Generator']:
        """
        Parses the commandline and returns the list of generator plugins.

        :param cmdline: the command-line to parse
        :type cmdline: str
        :param available_generators: the generators to use for parsing
        :type available_generators: dict
        :return: the list of generator plugins
        :rtype: list
        """
        generator_args = split_args(split_cmdline(cmdline), list(available_generators.keys()))
        return args_to_objects(generator_args, available_generators)


class SingleVariableGenerator(Generator, abc.ABC):
    """
    Ancestor for generators that output just a single variable.
    """

    def __init__(self, var_name: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the generator.

        :param var_name: the variable name
        :type var_name: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.var_name = var_name

    @abc.abstractmethod
    def _default_var_name(self) -> str:
        """
        Returns the default variable name.

        :return: the default name
        :rtype: str
        """
        raise NotImplementedError()

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-n", "--var_name", type=str, metavar="NAME", default=self._default_var_name(), help="The name of the variable", required=(self._default_var_name() is None))
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.var_name = ns.var_name

    def _check(self) -> Optional[str]:
        """
        Hook method for performing checks.

        :return: None if checks successful, otherwise error message
        :rtype: str
        """
        result = super()._check()

        if self.var_name is None:
            result = "No variable name provided!"

        return result


def test_generator(generator: Union[str, List[str]], generators_plugins: Dict[str, Plugin]):
    """
    Parses/executes the generator and then outputs the generated variables.

    :param generator: the generator command-line to use for generating variable values
    :type generator: str
    :param generators_plugins: the available generators
    :type generators_plugins: dict
    """
    vars_list = compile_generator_vars_list(generator, generators_plugins)
    for vars_ in vars_list:
        print(vars_)


def compile_generator_vars_list(generator: Union[str, List[str]], generator_plugins: Dict[str, Plugin]) -> List[Dict[str, str]]:
    """
    Generates the list of variable dictionaries to execute the pipeline with.

    :param generator: the generator command-line(s) to use for generating variable values to be expanded in the pipeline template
    :type generator: str or list
    :param generator_plugins: the available generator plugins (name -> plugin)
    :type generator_plugins: dict
    :return: the list of variables to execute the pipeline with
    :rtype: list
    """
    # only one item left in list?
    if isinstance(generator, list) and (len(generator) == 1):
        generator = generator[0]

    if isinstance(generator, str):
        generator_obj = Generator.parse_generator(generator, generator_plugins)
        result = generator_obj.generate()
    else:
        gen = generator.pop()
        last_list = compile_generator_vars_list(gen, generator_plugins)
        rest_list = compile_generator_vars_list(generator, generator_plugins)
        result = []
        for last in last_list:
            for rest in rest_list:
                combined = dict()
                combined.update(last)
                combined.update(rest)
                result.append(combined)

    return result


def perform_generator_test(env_var: Optional[str], args: List[str], prog: str, description: Optional[str],
                           generators: Dict[str, Plugin], logger: logging.Logger):
    """
    The main method for parsing command-line arguments.

    :param env_var: the environment variable for the logging level, can be None
    :type env_var: str
    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    :param prog: the name of the executable
    :type prog: str
    :param description: the description for the executable, uses default if None; list of available generators gets automatically appended
    :type description: str
    :param generators: the available generators
    :type generators: dict
    :param logger: the logger instance to use
    :type logger: logging.Logger
    """
    init_logging(env_var=env_var)
    if description is None:
        description = "Tool for testing generators by outputting the generated variables and their associated values."
    description += " Available generators: " + ", ".join(sorted(list(generators.keys())))
    parser = argparse.ArgumentParser(prog=prog, description=description, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-g", "--exec_generator", metavar="GENERATOR", help="The generator plugin(s) to use, incl. their options. Flag needs to be specified for each generator.", default=None, type=str, required=True, action="append")
    add_logging_level(parser)
    parsed = parser.parse_args(args=args)
    set_logging_level(logger, parsed.logging_level)
    test_generator(parsed.exec_generator, generators)



import argparse
import email
import imaplib
import os
import re
from time import sleep
from typing import List, Iterable

from email.header import decode_header
from dotenv import load_dotenv
from wai.logging import LOGGING_WARNING

from kasperl.api import Reader
from seppl.io import InfiniteReader
from seppl.placeholders import placeholder_list, PlaceholderSupporter, add_placeholder


IMAP_HOST = "IMAP_HOST"
IMAP_PORT = "IMAP_PORT"
IMAP_USER = "IMAP_USER"
IMAP_PW = "IMAP_PW"
IMAP_ENVS = [
    IMAP_HOST,
    IMAP_PORT,
    IMAP_USER,
    IMAP_PW,
]

DEFAULT_POLL_WAIT = 30.0

DEFAULT_POLL_WAIT_SLOW = 180.0

DEFAULT_POLL_COUNT = 10


MESSAGE_ACTION_NONE = "none"
MESSAGE_ACTION_READ = "read"
MESSAGE_ACTION_DELETE = "delete"
MESSAGE_ACTIONS = [
    MESSAGE_ACTION_NONE,
    MESSAGE_ACTION_READ,
    MESSAGE_ACTION_DELETE,
]


class GetEmail(Reader, InfiniteReader, PlaceholderSupporter):

    def __init__(self, dotenv_path: str = None, folder: str = None, only_unseen: bool = None, action: str = None,
                 regexp: str = None, output_dir: str = None, poll_wait: float = None,
                 poll_count: int = None, poll_wait_slow: float = None, max_poll: int = None,
                 from_placeholder: str = None, subject_placeholder: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param dotenv_path: the path to the .env file to use for initializing the IMAP environment vars
        :type dotenv_path: str
        :param folder: the folder to obtain the emails from, default is INBOX
        :type folder: str
        :param only_unseen: whether to list only unseen emails
        :type only_unseen: bool
        :param action: the action to apply to the retrieved emails
        :type action: str
        :param regexp: the regular expression that the attachments must match, ignored if None
        :type regexp: str
        :param output_dir: the path of the text file to load
        :type output_dir: str
        :param poll_wait: the seconds to wait between polls
        :type poll_wait: float
        :param poll_count: the number of times to use poll_wait before switching to poll_wait_slow
        :type poll_count: int
        :param poll_wait_slow: the seconds to wait between polls (during downtime)
        :type poll_wait_slow: float
        :param max_poll: the number of times to poll (<= 0 for infinite)
        :type max_poll: int
        :param from_placeholder: the placeholder name for storing the FROM email address under
        :type from_placeholder: str
        :param subject_placeholder: the placeholder name for storing the SUBJECT under
        :type subject_placeholder: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.dotenv_path = dotenv_path
        self.folder = folder
        self.only_unseen = only_unseen
        self.action = action
        self.regexp = regexp
        self.output_dir = output_dir
        self.poll_wait = poll_wait
        self.poll_count = poll_count
        self.poll_wait_slow = poll_wait_slow
        self.max_poll = max_poll
        self.from_placeholder = from_placeholder
        self.subject_placeholder = subject_placeholder
        self._dotenv_loaded = False
        self._server = None
        self._polled = None
        self._empty_poll_count = None
        self._slow_message_displayed = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "get-email"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Retrieves emails from the specified IMAP folder, saves the attachments " \
               "in the specified folder and forwards the file names of the saved attachments as list. " \
               "If the number of polls without any new messages reaches the 'poll_count' threshold, " \
               "the polling switches from the 'poll_wait' interval to 'poll_wait_slow'. It will " \
               "automatically reset the next time a new message is encountered."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-d", "--dotenv_path", metavar="FILE", type=str, help="The .env file to load the IMAP environment variables form (" + "|".join(IMAP_ENVS) + "); tries to load .env from the current directory if not specified; " + placeholder_list(obj=self), required=False, default=None)
        parser.add_argument("-f", "--folder", metavar="FOLDER", type=str, help="The IMAP folder to obtain emails from.", required=False, default="INBOX")
        parser.add_argument("-u", "--only_unseen", action="store_true", help="Whether to only retrieve unseen/new emails.", required=False)
        parser.add_argument("-a", "--action", choices=MESSAGE_ACTIONS, default=MESSAGE_ACTION_NONE, help="The action to apply to the retrieved emails.", required=False)
        parser.add_argument("-r", "--regexp", metavar="REGEXP", type=str, help="The regular expression that the attachment file names must match.", required=False, default=None)
        parser.add_argument("-o", "--output_dir", metavar="DIR", type=str, help="The directory to store the attachments in; " + placeholder_list(obj=self), required=True)
        parser.add_argument("-w", "--poll_wait", metavar="SECONDS", type=float, help="The poll interval in seconds", required=False, default=DEFAULT_POLL_WAIT)
        parser.add_argument("-W", "--poll_wait_slow", metavar="SECONDS", type=float, help="The poll interval in seconds during slow operation", required=False, default=DEFAULT_POLL_WAIT_SLOW)
        parser.add_argument("-c", "--poll_count", metavar="THRESHOLD", type=int, help="The maximum number of 'empty' polls that are allowed before switching from 'poll_wait' to 'poll_wait_slow'.", required=False, default=DEFAULT_POLL_COUNT)
        parser.add_argument("-m", "--max_poll", metavar="MAX", type=int, help="The maximum number of times to poll the folder; use <= for infinite polling.")
        parser.add_argument("-F", "--from_placeholder", metavar="PLACEHOLDER", type=str, help="The optional placeholder name to store the FROM email address under, without curly brackets.", required=False, default=None)
        parser.add_argument("-S", "--subject_placeholder", metavar="PLACEHOLDER", type=str, help="The optional placeholder name to store the SUBJECT under, without curly brackets.", required=False, default=None)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.dotenv_path = ns.dotenv_path
        self.folder = ns.folder
        self.only_unseen = ns.only_unseen
        self.action = ns.action
        self.regexp = ns.regexp
        self.output_dir = ns.output_dir
        self.poll_wait = ns.poll_wait
        self.poll_wait_slow = ns.poll_wait_slow
        self.poll_count = ns.poll_count
        self.max_poll = ns.max_poll
        self.from_placeholder = ns.from_placeholder
        self.subject_placeholder = ns.subject_placeholder
        self._dotenv_loaded = False
        self._server = None

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [str]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.folder is None:
            self.folder = "INBOX"
        if self.only_unseen is None:
            self.only_unseen = False
        if self.action is None:
            self.action = MESSAGE_ACTION_NONE
        if self.regexp is None:
            self.regexp = ""
        if self.output_dir is None:
            raise Exception("No output directory specified!")
        if self.poll_wait is None:
            self.poll_wait = DEFAULT_POLL_WAIT
        if self.poll_wait < 0:
            self.logger().warning("Invalid poll wait '%s', falling back to default of '%s'!" % (str(self.poll_wait), str(DEFAULT_POLL_WAIT)))
            self.poll_wait = DEFAULT_POLL_WAIT
        if self.poll_wait_slow is None:
            self.poll_wait_slow = DEFAULT_POLL_WAIT_SLOW
        if self.poll_wait_slow < 0:
            self.logger().warning("Invalid (slow) poll wait '%s', falling back to default of '%s'!" % (str(self.poll_wait_slow), str(DEFAULT_POLL_WAIT_SLOW)))
            self.poll_wait_slow = DEFAULT_POLL_WAIT_SLOW
        if self.poll_wait_slow < self.poll_wait:
            self.logger().warning("Slow poll wait is lower than normal poll wait, adjusting to '%s' for both!" % str(self.poll_wait))
            self.poll_wait_slow = self.poll_wait
        if self.poll_count is None:
            self.poll_count = DEFAULT_POLL_COUNT
        if self.max_poll is None:
            self.max_poll = -1
        self._polled = 0
        self._empty_poll_count = 0
        self._slow_message_displayed = False

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        output_dir = self.session.expand_placeholders(self.output_dir)
        if not os.path.exists(output_dir):
            raise Exception("Output directory does not exist: %s" % output_dir)

        # initialize environment variables
        if not self._dotenv_loaded:
            self._dotenv_loaded = True
            if self.dotenv_path is None:
                path = self.session.expand_placeholders("{CWD}/.env")
                self.logger().info("Loading .env from current dir: %s" % path)
                load_dotenv(path)
            else:
                path = self.session.expand_placeholders(self.dotenv_path)
                self.logger().info("Loading .env: %s" % path)
                load_dotenv(dotenv_path=path)

        poll_wait = self.poll_wait
        # slow down polling?
        if self._empty_poll_count > self.poll_count:
            if not self._slow_message_displayed:
                self._slow_message_displayed = True
                self.logger().info("Number of empty polls reached threshold (%s), switching to slow poll (%ss)." % (str(self.poll_count), str(self.poll_wait_slow)))
            poll_wait = self.poll_wait_slow
        # wait before polling?
        if poll_wait > 0:
            self.logger().info("Waiting for %s seconds before polling" % str(poll_wait))
            sleep(poll_wait)

        try:
            # connect
            if self._server is None:
                self.logger().info("Connection to IMAP host: %s" % os.getenv(IMAP_HOST) + ":" + os.getenv(IMAP_PORT))
                self._server = imaplib.IMAP4_SSL(os.getenv(IMAP_HOST), port=int(os.getenv(IMAP_PORT)))
                self.logger().info("Logging in...")
                self._server.login(os.getenv(IMAP_USER), os.getenv(IMAP_PW))

            self._polled += 1

            # select folder
            self.logger().info("Selecting folder: %s" % self.folder)
            self._server.select(mailbox=self.folder, readonly=False)

            # search messages
            self.logger().info("Looking for emails...")
            if self.only_unseen:
                status, messages = self._server.search(None, 'UNSEEN')
            else:
                status, messages = self._server.search(None, 'ALL')

            if status != 'OK':
                self.logger().warning("Failed to search emails: %s" % status)
            else:
                if isinstance(messages[0], bytes):
                    msg_list = messages[0].decode("utf-8")
                else:
                    msg_list = messages[0]
                if len(msg_list) == 0:
                    self.logger().info("No messages found.")
                    self._empty_poll_count += 1
                    return None

                # reset counter
                if self._empty_poll_count > self.poll_count:
                    self.logger().info("Switching back to normal poll wait (%ss)." % str(self.poll_wait))
                self._empty_poll_count = 0
                self._slow_message_displayed = False

                msg_ids = msg_list.split(' ')
                self.logger().info("# of messages found: %d" % len(msg_ids))
                for msg_id in msg_ids:
                    self.logger().info("Fetching message: %s" % msg_id)
                    status, msg_data = self._server.fetch(msg_id, '(RFC822)')
                    files = []
                    from_ = None
                    subject_ = None
                    for response_part in msg_data:
                        if isinstance(response_part, tuple):
                            msg = email.message_from_bytes(response_part[1])
                            subject_, _ = decode_header(msg["Subject"])[0]
                            from_, _ = decode_header(msg["From"])[0]
                            if not msg.is_multipart():
                                self.logger().info("No attachments, skipping!")
                            else:
                                for part in msg.walk():
                                    content_disposition = str(part.get("Content-Disposition"))
                                    if "attachment" in content_disposition:
                                        # download attachment
                                        filename = part.get_filename()
                                        if filename is not None:
                                            # does the file name match the regexp?
                                            if len(self.regexp) > 0:
                                                if not re.match(self.regexp, filename):
                                                    self.logger().info("Skipping attachment: %s" % filename)
                                                    continue
                                            output_file = os.path.join(self.session.expand_placeholders(self.output_dir), filename)
                                            self.logger().info("Saving attachment to: %s" % output_file)
                                            with open(output_file, "wb") as fp:
                                                fp.write(part.get_payload(decode=True))
                                            files.append(output_file)

                                # message action?
                                if self.action == MESSAGE_ACTION_READ:
                                    self.logger().info("Marking message as read: %s" % msg_id)
                                    status, data = self._server.uid('store', msg_id, '+FLAGS', '(\\Seen)')
                                    if status != 'OK':
                                        self.logger().warning("Failed to mark message as read: %s" % msg_id)
                                elif self.action == MESSAGE_ACTION_DELETE:
                                    self.logger().info("Deleting message: %s" % msg_id)
                                    status, data = self._server.store(msg_id, '+FLAGS', '\\Deleted')
                                    if status != 'OK':
                                        self.logger().warning("Failed to delete message: %s" % msg_id)
                                elif self.action == MESSAGE_ACTION_NONE:
                                    pass
                                else:
                                    self.logger().warning("Unsupported message action: %s" % self.action)

                    # forward file names of downloaded attachments
                    if len(files) > 0:
                        if (self.from_placeholder is not None) and (from_ is not None):
                            self.logger().info("Setting placeholder '%s' to: %s" % (self.from_placeholder, from_))
                            add_placeholder(self.from_placeholder, "from get-email", False, lambda i: from_)
                        if (self.subject_placeholder is not None) and (subject_ is not None):
                            self.logger().info("Setting placeholder '%s' to: %s" % (self.subject_placeholder, subject_))
                            add_placeholder(self.subject_placeholder, "from get-email", False, lambda i: subject_)
                        yield files
        except:
            self.logger().error("Failed to get emails!", exc_info=True)
            self._server = None

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return (self.max_poll > 0) and (self._polled >= self.max_poll)

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()
        if self.max_poll > 0:
            self.logger().info("# of polls: %s" % str(self._polled))
        if self._server is not None:
            try:
                self._server.close()
            except:
                pass
            self._server = None



import argparse
from typing import List

from seppl import AnyData
from seppl.io import BatchFilter
from seppl.placeholders import add_placeholder, InputBasedPlaceholderSupporter, placeholder_list
from wai.logging import LOGGING_WARNING


class SetPlaceholder(BatchFilter, InputBasedPlaceholderSupporter):

    def __init__(self, placeholder: str = None, value: str = None, use_current: bool = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param placeholder: the name of the placeholder (without curly brackets)
        :type placeholder: str
        :param value: the value of the placeholder
        :type value: str
        :param use_current: whether to use the current data passing through instead of the value
        :type use_current: bool
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.placeholder = placeholder
        self.value = value
        self.use_current = use_current

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "set-placeholder"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Sets the placeholder to the specified value when data passes through. The value can contain other placeholders, which get expanded each time data passes through. Can use the data passing through instead of specified value as well."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-p", "--placeholder", type=str, help="The name of the placeholder, without curly brackets.", default=None, required=True)
        parser.add_argument("-v", "--value", type=str, help="The value of the placeholder, may contain other placeholders. " + placeholder_list(obj=self), default=None, required=False)
        parser.add_argument("-u", "--use_current", action="store_true", help="Whether to use the data passing through instead of the specified value.", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.placeholder = ns.placeholder
        self.value = ns.value
        self.use_current = ns.use_current

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.placeholder is None:
            raise Exception("No placeholder name provided!")
        if (not self.use_current) and (self.value is None):
            raise Exception("No placeholder value provided!")

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        if self.use_current:
            value = str(data)
        else:
            value = self.value
        value = self.session.expand_placeholders(value)
        self.logger().info("%s -> %s" % (self.placeholder, value))
        add_placeholder(self.placeholder, "no description", False, lambda i: value)
        return data



import argparse
import os
from typing import List

from wai.logging import LOGGING_WARNING

from seppl import AnyData
from seppl.placeholders import InputBasedPlaceholderSupporter, placeholder_list
from kasperl.api import make_list, StreamWriter


class TextFileWriter(StreamWriter, InputBasedPlaceholderSupporter):

    def __init__(self, output_file: str = None, append: bool = None, delete_on_initialize: bool = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the writer.

        :param output_file: the file to write to
        :type output_file: str
        :param append: whether to append the file rather than overwrite
        :type append: bool
        :param delete_on_initialize: whether to delete an existing file when initializing the writer
        :type delete_on_initialize: bool
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.output_file = output_file
        self.append = append
        self.delete_on_initialize = delete_on_initialize

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-text-file"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Stores the incoming data in the specified text file."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output_file", metavar="FILE", type=str, help="The file to write the data to; " + placeholder_list(obj=self), required=True)
        parser.add_argument("-a", "--append", action="store_true", help="Whether to append the file rather than overwrite it.")
        parser.add_argument("-d", "--delete_on_initialize", action="store_true", help="Whether to remove any existing file when initializing the writer.")
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_file = ns.output_file
        self.append = ns.append
        self.delete_on_initialize = ns.delete_on_initialize

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        output_file = self.session.expand_placeholders(self.output_file)
        if os.path.exists(output_file) and os.path.isfile(output_file):
            self.logger().info("Deleting during initialization: %s" % output_file)
            os.remove(output_file)

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        for item in make_list(data):
            output_file = self.session.expand_placeholders(self.output_file)
            if self.append:
                with open(output_file, "a") as fp:
                    fp.write(str(item))
                    fp.write("\n")
            else:
                with open(output_file, "w") as fp:
                    fp.write(str(item))
                    fp.write("\n")



import argparse
from typing import List, Any

from wai.logging import LOGGING_WARNING

from seppl import MetaDataHandler, AnyData, METADATA_TYPES, METADATA_TYPE_STRING, METADATA_TYPE_BOOL, METADATA_TYPE_NUMERIC
from seppl.io import BatchFilter
from kasperl.api import make_list, flatten_list


class SetMetaData(BatchFilter):
    """
    Sets the specified key/value pair in the meta-data.
    """

    def __init__(self, field: str = None, value=None, as_type: str = None, use_current: bool = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param field: the name of the meta-data field to set
        :type field: str
        :param value: the value to set
        :param as_type: the type to convert the value to
        :type as_type: str
        :param use_current: whether to use the current data passing through instead of the value
        :type use_current: bool
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)

        self.field = field
        self.value = value
        self.as_type = as_type
        self.use_current = use_current
        self._value = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "set-metadata"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Sets the specified key/value pair in the meta-data. Can use the data passing through instead of the specified value as well."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-f", "--field", type=str, help="The meta-data field to use in the comparison", required=True)
        parser.add_argument("-v", "--value", type=str, help="The value to use in the comparison", required=False)
        parser.add_argument("-t", "--as_type", choices=METADATA_TYPES, default=METADATA_TYPE_STRING, help="How to interpret the value")
        parser.add_argument("-u", "--use_current", action="store_true", help="Whether to use the data passing through instead of the specified value.", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.field = ns.field
        self.value = ns.value
        self.as_type = ns.as_type
        self.use_current = ns.use_current

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.field is None:
            raise Exception("No meta-data field provided!")
        if (not self.use_current) and (self.value is None):
            raise Exception("No value provided to compare with!")
        if self.as_type is None:
            self.as_type = METADATA_TYPE_STRING
        if not self.use_current:
            self._value = self._to_type(self.value)

    def _to_type(self, value: str) -> Any:
        """
        Turns the value into the specified type.

        :param value: the value to process
        :type value: str
        :return: the converted value
        """
        if self.as_type == METADATA_TYPE_STRING:
            return str(value)
        elif self.as_type == METADATA_TYPE_BOOL:
            return str(value).lower() == "true"
        elif self.as_type == METADATA_TYPE_NUMERIC:
            return float(value)
        else:
            raise Exception("Unhandled meta-data type: %s" % self.as_type)

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        if self.use_current:
            value = self._to_type(data)
        else:
            value = self._value

        for item in make_list(data):
            result.append(item)
            meta = None
            if isinstance(item, MetaDataHandler):
                if item.has_metadata():
                    meta = item.get_metadata()

            if meta is None:
                self.logger().info("No meta-data, ignoring")
            else:
                self.logger().info("Setting meta-data: %s=%s" % (self.field, str(value)))
                meta[self.field] = value

        return flatten_list(result)



import abc
import argparse
from typing import List, Dict

from wai.logging import LOGGING_WARNING

from kasperl.api import make_list, compare_values, \
    COMPARISONS_EXT, COMPARISON_EQUAL, COMPARISON_CONTAINS, COMPARISON_MATCHES, COMPARISON_EXT_HELP, \
    PIPELINE_FORMATS, PIPELINE_FORMAT_CMDLINE, load_pipeline
from seppl import split_args, Plugin, AnyData, Initializable, MetaDataHandler, init_initializable
from seppl.io import Writer, BatchWriter, StreamWriter, BatchFilter, MultiFilter


class Tee(BatchFilter, abc.ABC):
    """
    Forwards the data coming through to the sub-flow.
    """

    def __init__(self, sub_flow: str = None, sub_flow_format: str = None,
                 field: str = None, comparison: str = COMPARISON_EQUAL, value=None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param sub_flow: the command-line/pipeline file with the filter(s)/writer to execute
        :type sub_flow: str
        :param sub_flow_format: the format the sub_flow is in
        :type sub_flow_format: str
        :param field: the name of the meta-data field to perform the comparison on
        :type field: str
        :param comparison: the comparison to perform
        :type comparison: str
        :param value: the value to compare with
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.sub_flow = sub_flow
        self.sub_flow_format = sub_flow_format
        self.field = field
        self.value = value
        self.comparison = comparison
        self._sub_flow = None
        self._filter = None
        self._writer = None
        self._data_buffer = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "tee"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Forwards the data passing through to the filter/writer defined as its sub-flow. " \
               "When supplying a meta-data field and a value, this can be turned into a conditional forwarding. " \
               "Performs the following comparison: METADATA_VALUE COMPARISON VALUE."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-f", "--sub_flow", type=str, default=None, help="The subflow to execute (filter(s)/writer).")
        parser.add_argument("-F", "--sub_flow_format", choices=PIPELINE_FORMATS, default=PIPELINE_FORMAT_CMDLINE, help="The format of the pipeline.")
        parser.add_argument("--field", type=str, help="The meta-data field to use in the comparison", default=None, required=False)
        parser.add_argument("--comparison", choices=COMPARISONS_EXT, default=COMPARISON_EQUAL, help="How to compare the value with the meta-data value; " + COMPARISON_EXT_HELP
                            + "; in case of '" + COMPARISON_CONTAINS + "' and '" + COMPARISON_MATCHES + "' the supplied value represents the substring to find/regexp to search with", required=False)
        parser.add_argument("--value", type=str, help="The value to use in the comparison", default=None, required=False)
        return parser

    @abc.abstractmethod
    def _available_filters(self) -> Dict[str, Plugin]:
        """
        Returns the available filters from the registry.

        :return: the filters
        :rtype: dict
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _available_writers(self) -> Dict[str, Plugin]:
        """
        Returns the available writers from the registry.

        :return: the writers
        :rtype: dict
        """
        raise NotImplementedError()

    def _parse_sub_flow(self) -> List[Plugin]:
        """
        Parses the command-line and returns the list of plugins it represents.
        Raises an exception in case of an invalid sub-flow.
        
        :return: the list of plugins
        :rtype: list
        """
        from seppl import args_to_objects

        # split command-line into valid plugin subsets
        valid = dict()
        valid.update(self._available_filters())
        valid.update(self._available_writers())
        pipeline = load_pipeline(self.sub_flow, self.sub_flow_format, logger=self.logger())
        args = split_args(pipeline, list(valid.keys()))
        return args_to_objects(args, valid, allow_global_options=False)

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.sub_flow = ns.sub_flow
        self.sub_flow_format = ns.sub_flow_format
        self.field = ns.field
        self.value = ns.value
        self.comparison = ns.comparison

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()

        if self.sub_flow is None:
            self._sub_flow = []
        else:
            self._sub_flow = self._parse_sub_flow()

        if len(self._sub_flow) > 0:
            self._filter = None
            self._writer = None
            filters = []
            for plugin in self._sub_flow:
                if isinstance(plugin, BatchFilter):
                    filters.append(plugin)
                elif isinstance(plugin, Writer):
                    self._writer = plugin
                    # no more plugins allowed beyond writer
                    break
            if len(filters) == 1:
                self._filter = filters[0]
            elif len(filters) > 1:
                self._filter = MultiFilter(filters=filters)
            if (self._writer is not None) and isinstance(self._writer, BatchWriter):
                self._data_buffer = []

        if self._filter is not None:
            self._filter.session = self.session
            if isinstance(self._filter, Initializable):
                init_initializable(self._filter, "filter")
        if self._writer is not None:
            self._writer.session = self.session
            if isinstance(self._writer, Initializable):
                init_initializable(self._writer, "writer")

        if (self.field is not None) and (self.value is None):
            raise Exception("No value provided to compare with!")

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        for item in make_list(data):
            # evaluate expression?
            meta = None
            if self.field is not None:
                if isinstance(item, MetaDataHandler):
                    if item.has_metadata():
                        meta = item.get_metadata()
            if meta is not None:
                v1 = meta[self.field]
                v2 = self.value
                comp_result = compare_values(v1, self.comparison, v2)
                comp = str(meta[self.field]) + " " + self.comparison + " " + str(self.value) + " = " + str(comp_result)
                self.logger().info("Field '%s': '%s'" % (self.field, comp))
                if not comp_result:
                    continue

            # filter data
            if self._filter is not None:
                item = self._filter.process(item)

            # write data
            if (item is not None) and (self._writer is not None):
                if isinstance(self._writer, BatchWriter):
                    self._data_buffer.append(item)
                elif isinstance(self._writer, StreamWriter):
                    self._writer.write_stream(item)
                else:
                    raise Exception("Unhandled type of writer: %s" % str(type(self._writer)))

        return data

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()

        # flush batch buffer
        if (self._data_buffer is not None) and isinstance(self._writer, BatchWriter):
            self.logger().info("Flushing data buffer...")
            self._writer.write_batch(self._data_buffer)
            self._data_buffer = None

        # finalize sub-flow
        if (self._filter is not None) and isinstance(self._filter, Initializable):
            self._filter.finalize()
        if (self._writer is not None) and isinstance(self._writer, Initializable):
            self._writer.finalize()



import argparse
from typing import List

from seppl import AnyData
from seppl.io import BatchFilter
from wai.logging import LOGGING_WARNING
from kasperl.api import flatten_list, make_list


class RecordWindow(BatchFilter):
    """
    Only lets records pass that match the defined window and step size.
    """

    def __init__(self, from_index: int = None, to_index: int = None, step: int = 1, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param from_index: the lower bound for the window (1-based, included), ignored if None
        :type from_index: int
        :param to_index: the upper bound for the window (1-based, included), ignored if None
        :type to_index: int
        :param step: the increment to use (at least 1)
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.from_index = from_index
        self.to_index = to_index
        self.step = step
        self._counter = 0

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "record-window"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Only lets records pass that match the defined window and step size."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-f", "--from_index", type=int, help="The 1-based lower bound of the window, ignored if not supplied.", default=None, required=False)
        parser.add_argument("-t", "--to_index", type=int, help="The 1-based upper bound of the window, ignored if not supplied.", default=None, required=False)
        parser.add_argument("-s", "--step", type=int, help="The increment to use (at least 1).", default=1, required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.from_index = ns.from_index
        self.to_index = ns.to_index
        self.step = ns.step

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._counter = 0
        if self.from_index is not None:
            if self.from_index < 1:
                raise Exception("From index must be at least 1, provided: %d" % self.from_index)
        if self.to_index is not None:
            if self.to_index < 1:
                raise Exception("To index must be at least 1, provided: %d" % self.to_index)
        if (self.from_index is not None) and (self.to_index is not None):
            if self.to_index < self.from_index:
                raise Exception("To index cannot be smaller than from index: from=%d to=%d" % (self.from_index, self.to_index))
        if self.step < 1:
            raise Exception("Step must be at least 1, provided: %d" % self.step)

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            self._counter += 1

            keep = True
            if self.from_index is not None:
                if self._counter < self.from_index:
                    keep = False
            if self.to_index is not None:
                if self._counter > self.to_index:
                    keep = False
            if self.step > 1:
                min_ = 1 if (self.from_index is None) else self.from_index
                if (self._counter - min_) % self.step > 0:
                    keep = False

            self.logger().debug("from=%s to=%s step=%s counter=%s -> %s"
                                % (str(self.from_index), str(self.to_index), str(self.step), str(self._counter), str(keep)))

            if keep:
                result.append(item)

        return flatten_list(result)



import argparse
import logging
import os.path
import sys
import traceback
from typing import List, Tuple, Optional, Dict

from wai.logging import set_logging_level, LOGGING_LEVELS, init_logging, LOGGING_WARNING

from kasperl.api import Session
from seppl import enumerate_plugins, is_help_requested, split_args, args_to_objects, Plugin, check_compatibility, \
    save_args
from seppl.io import Reader, BatchFilter, MultiFilter, Writer, execute
from seppl.placeholders import load_user_defined_placeholders, expand_placeholders
from ._exec import load_pipeline, PIPELINE_FORMAT_FILE
from ._help import CommandlineParameter, params_to_short, param_to_help, params_to_parser

DEFAULT_UPDATE_INTERVAL = 1000

PARAM_LOAD_PIPELINE = "--load_pipeline"
PARAM_DUMP_PIPELINE = "--dump_pipeline"


def _default_params() -> List[CommandlineParameter]:
    """
    Returns the default options to use.

    :return: the list of option definitions
    :rtype: list
    """
    return [
        CommandlineParameter(short_opt="-h", long_opt="--help", help="Show basic help message and exit.", action="store_true", is_help=True),
        CommandlineParameter(long_opt="--help-all", help="Show basic help message plus help on all plugins and exit.", action="store_true", is_help=True),
        CommandlineParameter(long_opt="--help-plugin", metavar="NAME", help="Show help message for plugin NAME and exit.", is_help=True),
        CommandlineParameter(short_opt="-u", long_opt="--update_interval", metavar="INTERVAL", help="Outputs the progress every INTERVAL records (default: %d)." % DEFAULT_UPDATE_INTERVAL, type=int, default=DEFAULT_UPDATE_INTERVAL),
        CommandlineParameter(short_opt="-l", long_opt="--logging_level", choices=LOGGING_LEVELS, help="The logging level to use (default: WARN).", default=LOGGING_WARNING),
        CommandlineParameter(short_opt="-b", long_opt="--force_batch", help="Processes the data in batches.", action="store_true"),
        CommandlineParameter(long_opt="--placeholders", metavar="FILE", help="The file with custom placeholders to load (format: key=value)."),
        CommandlineParameter(long_opt=PARAM_LOAD_PIPELINE, metavar="FILE", help="The file to load the pipeline command from."),
        CommandlineParameter(long_opt=PARAM_DUMP_PIPELINE, metavar="FILE", help="The file to dump the pipeline command in."),
    ]


def print_conversion_usage(prog: str, description: str,
                           readers: Dict[str, Plugin], filters: Dict[str, Plugin], writers: Dict[str, Plugin],
                           aliases: List[str] = None, plugin_details: bool = False, generate_plugin_usage=None,
                           additional_params: Optional[List[CommandlineParameter]] = None):
    """
    Prints the program usage to stdout.
    Ensure global options are in sync with parser in parse_args method below.

    :param prog: the conversion executable
    :type prog: str
    :param description: the description of the executable
    :type description: str
    :param readers: the available readers
    :type readers: dict
    :param filters: the available filters
    :type filters: dict
    :param writers: the available writers
    :type writers: dict
    :param aliases: the list of aliases in the registry
    :type aliases: list
    :param plugin_details: whether to output the plugin details as well
    :type plugin_details: bool
    :param generate_plugin_usage: the method for generating the plugin usage
    :param additional_params: the additional parameters for the parser
    :type additional_params: list
    """
    params = _default_params()
    if additional_params is not None:
        params.extend(additional_params)
    print(params_to_short(prog, params))
    print()
    print(description)
    print()
    print("readers (%d):\n" % len(readers) + enumerate_plugins(readers.keys(), aliases=aliases, prefix="   "))
    print("filters (%d):\n" % len(filters) + enumerate_plugins(filters.keys(), aliases=aliases, prefix="   "))
    print("writers (%d):\n" % len(writers) + enumerate_plugins(writers.keys(), aliases=aliases, prefix="   "))
    print()
    print("options:")
    for param in params:
        print(param_to_help(param))
    print()
    if plugin_details and (generate_plugin_usage is not None):
        all_plugins = dict()
        all_plugins.update(readers)
        all_plugins.update(filters)
        all_plugins.update(writers)
        for plugin in sorted(all_plugins.keys()):
            generate_plugin_usage(plugin)


def parse_conversion_args(args: List[str], prog: str, description: str,
                          readers: Dict[str, Plugin], filters: Dict[str, Plugin], writers: Dict[str, Plugin],
                          aliases: List[str] = None, require_reader: bool = True, require_writer: bool = True,
                          generate_plugin_usage=None, exit_on_help: bool = True,
                          additional_params: Optional[List[CommandlineParameter]] = None) -> Optional[Tuple[Optional[Reader], Optional[BatchFilter], Optional[Writer], Session]]:
    """
    Parses the arguments for the conversion tool.

    :param args: the arguments to parse
    :type args: list
    :param prog: the conversion executable
    :type prog: str
    :param description: the description of the executable
    :type description: str
    :param readers: the available readers
    :type readers: dict
    :param filters: the available filters
    :type filters: dict
    :param writers: the available writers
    :type writers: dict
    :param aliases: the list of aliases in the registry
    :type aliases: list
    :param require_reader: whether a reader is required
    :type require_reader: bool
    :param require_writer: whether a writer is required
    :type require_writer: bool
    :param generate_plugin_usage: the method for generating the plugin usage
    :param exit_on_help: whether to perform a sys.exit(0) when help gets requested or just return None
    :type exit_on_help: bool
    :param additional_params: the additional parameters for the parser
    :type additional_params: list
    :return: tuple of (reader, filter, writer, session), the filter can be None
    :rtype: tuple
    """
    logger = logging.getLogger(prog)

    # prepare plugins
    partial = False
    all_plugins = dict()
    all_plugins.update(readers)
    all_plugins.update(filters)
    all_plugins.update(writers)
    handlers = list(all_plugins.keys())

    # help requested?
    help_requested, plugin_details, plugin_name = is_help_requested(args, handlers=handlers, partial=partial)
    if help_requested:
        if (plugin_name is not None) and (generate_plugin_usage is not None):
            generate_plugin_usage(plugin_name)
        else:
            print_conversion_usage(prog, description, readers, filters, writers,
                                   aliases=aliases, generate_plugin_usage=generate_plugin_usage,
                                   plugin_details=plugin_details, additional_params=additional_params)
        if exit_on_help:
            sys.exit(0)
        else:
            return None

    # load_pipeline from file?
    if PARAM_LOAD_PIPELINE in args:
        idx = args.index(PARAM_LOAD_PIPELINE)
        pipeline_file = args[idx+1]
        pipeline_args = load_pipeline(pipeline_file, pipeline_format=PIPELINE_FORMAT_FILE,
                                      remove_convert_prog=True, convert_prog=prog, logger=logger)
        # remove "load pipeline" args
        del args[idx]
        del args[idx]
        # append pipeline args from file
        args.extend(pipeline_args)

    # parse arguments
    parsed = split_args(args, handlers, partial=partial)
    plugins = args_to_objects(parsed, all_plugins, allow_global_options=True)
    reader = None
    writer = None
    filters = []
    for plugin in plugins:
        if isinstance(plugin, Reader):
            if reader is None:
                reader = plugin
                continue
            else:
                raise Exception("Only one reader can be defined!")

        if isinstance(plugin, BatchFilter):
            filters.append(plugin)
            continue

        if isinstance(plugin, Writer):
            if writer is None:
                writer = plugin
                continue
            else:
                raise Exception("Only one writer can be defined!")

        raise Exception("Unhandled plugin type: %s" % str(type(plugin)))

    # checks whether valid pipeline
    if (reader is None) and require_reader:
        raise Exception("No reader defined!")
    if (writer is None) and require_writer:
        raise Exception("No writer defined!")
    if len(filters) == 0:
        filter_ = None
    elif len(filters) == 1:
        filter_ = filters[0]
    else:
        filter_ = MultiFilter(filters=filters)

    # check compatibility
    if writer is not None:
        if filter_ is not None:
            check_compatibility([reader, filter_, writer])
        else:
            check_compatibility([reader, writer])

    # global options?
    # see print_usage() method above
    params = _default_params()
    if additional_params is not None:
        params.extend(additional_params)
    parser = argparse.ArgumentParser()
    params_to_parser(parser, params)
    session = Session(options=parser.parse_args(parsed[""] if ("" in parsed) else []),
                      logger=logger)
    set_logging_level(session.logger, session.options.logging_level)
    if session.options.placeholders is not None:
        if not os.path.exists(session.options.placeholders):
            session.logger.error("Placeholder file not found: %s" % session.options.placeholders)
        else:
            session.logger.info("Loading custom placeholders from: %s" % session.options.placeholders)
            load_user_defined_placeholders(session.options.placeholders)
    if session.options.dump_pipeline is not None:
        # remove "--dump_pipeline ..." args
        dump_args = args[:]
        idx = dump_args.index(PARAM_DUMP_PIPELINE)
        del dump_args[idx]
        del dump_args[idx]
        # save arguments
        pipeline_path = expand_placeholders(session.options.dump_pipeline)
        save_args(dump_args, pipeline_path, handlers=handlers, prog=prog, logger=logger)

    return reader, filter_, writer, session


def perform_conversion(env_var: Optional[str], args: List[str], prog: str, description: str,
                       readers: Dict[str, Plugin], filters: Dict[str, Plugin], writers: Dict[str, Plugin],
                       aliases: List[str] = None, require_reader: bool = True, require_writer: bool = True,
                       generate_plugin_usage=None, additional_params: Optional[List[CommandlineParameter]] = None,
                       pre_initialize=None, post_finalize=None):
    """
    Parses the command-line arguments and performs the conversion.

    :param env_var: the environment variable to obtain the logging level from, can be None
    :type env_var: str
    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    :param args: the arguments to parse
    :type args: list
    :param prog: the conversion executable
    :type prog: str
    :param description: the description of the executable
    :type description: str
    :param readers: the available readers
    :type readers: dict
    :param filters: the available filters
    :type filters: dict
    :param writers: the available writers
    :type writers: dict
    :param aliases: the list of aliases in the registry
    :type aliases: list
    :param require_reader: whether a reader is required
    :type require_reader: bool
    :param require_writer: whether a writer is required
    :type require_writer: bool
    :param generate_plugin_usage: the method for generating the plugin usage
    :param additional_params: the additional parameters for the parser
    :type additional_params: list
    :param pre_initialize: optional method to execute before the plugins get initialized, takes the session object as only parameter
    :param post_finalize: optional method to execute after the plugins have been finalized, takes the session object as only parameter
    """
    init_logging(env_var=env_var)
    _args = sys.argv[1:] if (args is None) else args
    try:
        reader, filter_, writer, session = parse_conversion_args(
            _args, prog, description, readers, filters, writers,
            aliases=aliases, require_reader=require_reader, require_writer=require_writer,
            generate_plugin_usage=generate_plugin_usage, additional_params=additional_params)
        session.logger.info("options: %s" % str(_args))
        execute(reader, filter_, writer, session, pre_initialize=pre_initialize, post_finalize=post_finalize)
    except Exception:
        traceback.print_exc()
        print("options: %s" % str(_args), file=sys.stderr)
        print_conversion_usage(
            prog, description,
            readers, filters, writers,
            generate_plugin_usage=generate_plugin_usage)
        sys.exit(1)



import argparse
import os
import re
from typing import List, Iterable

from wai.logging import LOGGING_WARNING

from kasperl.api import Reader
from seppl.placeholders import placeholder_list, PlaceholderSupporter


class ListFiles(Reader, PlaceholderSupporter):

    def __init__(self, input_dir: str = None, regexp: str = None, as_list: bool = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param input_dir: the directory to list files from
        :type input_dir: str
        :param regexp: the regular expression that the files must match
        :type regexp: str
        :param as_list: whether to output the files as a list rather than one by one
        :type as_list: bool
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.input_dir = input_dir
        self.regexp = regexp
        self.as_list = as_list
        self._finished = False

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "list-files"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Lists files in the specified directory and forwards them."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input_dir", metavar="DIR", type=str, help="The directory to list the files in; " + placeholder_list(obj=self), required=True)
        parser.add_argument("-r", "--regexp", metavar="REGEXP", type=str, help="The regular expression that the files must match.", required=False, default=".*")
        parser.add_argument("--as_list", action="store_true", help="Whether to forward the files as a list or one by one.", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.input_dir = ns.input_dir
        self.regexp = ns.regexp
        self.as_list = ns.as_list

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        if self.as_list:
            return [List[str]]
        else:
            return [str]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.input_dir is None:
            raise Exception("No input_dir specified!")
        if self.regexp is None:
            self.regexp = ".*"
        if self.as_list is None:
            self.as_list = False

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        input_dir = self.session.expand_placeholders(self.input_dir)
        if not os.path.exists(input_dir):
            raise Exception("input_dir does not exist: %s" % input_dir)
        if not os.path.isdir(input_dir):
            raise Exception("Not a directory: %s" % input_dir)
        self.logger().info("Listing files in: %s" % input_dir)
        self._finished = False
        files = []
        for f in os.listdir(input_dir):
            full = os.path.join(input_dir, f)
            if os.path.isdir(full):
                continue
            if re.match(self.regexp, f):
                files.append(full)
        if self.as_list:
            yield files
        else:
            for f in files:
                yield f
        self._finished = True

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return self._finished



from ._console import ConsoleWriter
from ._delete_files import DeleteFiles
from ._metadata import MetaDataWriter
from ._send_email import SendEmail, SMTP_ENVS, SMTP_HOST, SMTP_PORT, SMTP_STARTTLS, SMTP_USER, SMTP_PW
from ._storage import StorageWriter
from ._text_file import TextFileWriter



import argparse
from typing import List

from seppl import AnyData
from seppl.io import BatchFilter
from wai.logging import LOGGING_WARNING
from kasperl.api import flatten_list, make_list


class MaxRecords(BatchFilter):
    """
    Suppresses records after the specified maximum number of records have passed through.
    """

    def __init__(self, max_records: int = -1, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param max_records: the maximum number of records to process <= 0 is unlimited
        :type max_records: int
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.max_records = max_records
        self._counter = 0

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "max-records"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Suppresses records after the specified maximum number of records have passed through."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-m", "--max_records", type=int, help="The maximum number number of records to let through before suppressing records.", default=-1, required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.max_records = ns.max_records

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._counter = 0

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            # maximum reached?
            if (self.max_records > 0) and (self._counter >= self.max_records):
                if self._counter == self.max_records:
                    self.logger().info("Maximum # of records reached: %d" % self.max_records)
                    self.session.stopped = True
                item = None

            self._counter += 1

            if item is not None:
                result.append(item)

        return flatten_list(result)



import logging
from typing import List

from seppl import AnyData
from seppl.io import BatchFilter
from wai.logging import LOGGING_WARNING


class PassThrough(BatchFilter):
    """
    Dummy, just passes through the data.
    """

    def __init__(self, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "passthrough"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Dummy, just passes through the data."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        if self.logger().isEnabledFor(logging.INFO):
            self.logger().info(str(data))
        return data



import abc
import argparse
from typing import Dict, List, Iterable

from wai.logging import LOGGING_WARNING

import seppl.io
from seppl import Initializable, Plugin, AnyData, init_initializable


class Reader(seppl.io.Reader, Initializable, abc.ABC):
    """
    Ancestor for dataset readers.
    """
    pass


def parse_reader(reader: str, available_readers: Dict[str, Plugin]) -> Reader:
    """
    Parses the command-line and instantiates the reader.

    :param reader: the command-line to parse
    :type reader: str
    :param available_readers: the available readers to use for parsing
    :type available_readers: dict
    :return: the reader instance
    :rtype: Reader
    """
    from seppl import split_args, split_cmdline, args_to_objects

    if reader is None:
        raise Exception("No reader command-line supplied!")
    valid = dict()
    valid.update(available_readers)
    args = split_args(split_cmdline(reader), list(valid.keys()))
    objs = args_to_objects(args, valid, allow_global_options=False)
    if len(objs) == 1:
        if isinstance(objs[0], Reader):
            result = objs[0]
        else:
            raise Exception("Expected instance of Reader but got: %s" % str(type(objs[0])))
    else:
        raise Exception("Expected to obtain one reader from '%s' but got %d instead!" % (reader, len(objs)))
    return result


def add_annotations_only_reader_param(parser: argparse.ArgumentParser):
    """
    Adds the --annotations_only parameter to the parser, as used by readers of type AnnotationOnlyReader.

    :param parser: the parser
    :type parser: argparse.ArgumentParser
    """
    parser.add_argument("--annotations_only", action="store_true", help="Reads only the annotations.")


class AnnotationsOnlyReader:
    """
    Mixin for reader that can read the annotations by itself.
    """
    pass


class MetaFileReader(Reader, abc.ABC):
    """
    Ancestor for classes that use a base reader for doing the actual reading
    of the files that the reader determined.
    """

    def __init__(self, base_reader: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param base_reader: the base reader to use (command-line)
        :type base_reader: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.base_reader = base_reader
        self._base_reader = None

    def _base_reader_help(self) -> str:
        """
        Returns the help string on the base reader for the argument parser.

        :return: the help string
        :rtype: str
        """
        return "The command-line of the reader for reading the files"

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-b", "--base_reader", type=str, help=self._base_reader_help(), required=True, default=None)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.base_reader = ns.base_reader

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        if self._base_reader is None:
            return [AnyData]
        else:
            return [self._base_reader.generates()]

    @abc.abstractmethod
    def _available_readers(self) -> Dict[str, Plugin]:
        """
        Return the available readers.

        :return: the reader plugins
        :rtype: dict
        """
        raise NotImplementedError()

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """

        super().initialize()

        if self.base_reader is None:
            raise Exception("No base reader defined!")

        # configure base reader
        self._base_reader = parse_reader(self.base_reader, self._available_readers())
        if not hasattr(self._base_reader, "source"):
            raise Exception("Reader does not have 'source' attribute: %s" % str(type(self._base_reader)))
        self._base_reader.session = self.session

    def _read_files(self, files: List[str]) -> Iterable:
        """
        Reads the files with the base reader and returns the results.

        :param files: the files to read
        :type files: list
        :return: the data that was generated
        """
        result = []
        self._base_reader.source = files
        if isinstance(self._base_reader, Initializable):
            init_initializable(self._base_reader, "reader", raise_again=True)
        while not self._base_reader.has_finished():
            for item in self._base_reader.read():
                result.append(item)
        self._base_reader.finalize()
        return result

    def finalize(self):
        """
        Finishes the reading, e.g., for closing files or databases.
        """
        super().finalize()
        if self._base_reader is not None:
            self._base_reader.finalize()
            self._base_reader = None



import argparse
import os
import shutil
from typing import List, Optional

from wai.logging import LOGGING_WARNING

from kasperl.api import make_list, flatten_list
from seppl.io import BatchFilter
from seppl.placeholders import InputBasedPlaceholderSupporter, placeholder_list


class CopyFiles(BatchFilter, InputBasedPlaceholderSupporter):
    """
    Copies the files it receives into the specified target directory and forwards the new files.
    """

    def __init__(self, target_dir: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param target_dir: the directory to copy the files to
        :type target_dir: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.target_dir = target_dir

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "copy-files"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Copies the files it receives into the specified target directory and forwards the new files."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [str, list]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [str, list]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-t", "--target_dir", metavar="DIR", type=str, help="The directory to copy the files to. " + placeholder_list(obj=self), required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.target_dir = ns.target_dir

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.target_dir is None:
            raise Exception("Target directory not specified!")

    def _copy(self, source: str, target_path: str) -> Optional[str]:
        """
        Copies the file to the target directory and returns the new path.
        Returns None in case it encounters directories.

        :param source: the path to copy
        :type source: str
        :param target_path: the directory to copy to
        :type target_path: str
        :return: the new path, None if input was a directory (gets skipped) or failed to copy
        :rtype: str or None
        """
        if not os.path.isfile(source):
            return None

        dest = os.path.join(target_path, os.path.basename(source))
        try:
            self.logger().info("Copying '%s' to '%s'..." % (source, dest))
            shutil.copy(source, dest)
            return dest
        except:
            self.logger().error("Failed to copy '%s' to '%s'!" % (source, dest))
        return None

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        # apply placeholders
        target_path = self.session.expand_placeholders(self.target_dir)
        if target_path != self.target_dir:
            self.logger().info("Expanded target dir: %s" % target_path)
        if not os.path.exists(target_path):
            raise Exception("Target directory does not exist: %s" % target_path)
        if not os.path.isdir(target_path):
            raise Exception("Target directory is not a directory: %s" % target_path)

        # copy files
        result = []
        for item in make_list(data):
            if isinstance(item, list):
                for subitem in item:
                    new_path = self._copy(subitem, target_path)
                    if new_path is not None:
                        result.append(new_path)
            else:
                new_path = self._copy(item, target_path)
                if new_path is not None:
                    result.append(new_path)

        return flatten_list(result)
